;--------------------------------------------------------------------------
;
;	MFM Writer
;
;	Write Sector using only software to control the pins of the AVR
;
;	When writing a sector we need to write the complete data segment
;
;	+--------+----+----+----------------------------+-----+---------+
;	|preamble|sync|DAM |             Data           | CRC |  write  |
;	|  12 x  | 3x | 1x |            512 x           | 2x  | splice  |
;	|  0x00  |0xA1|0xFB|            user            | CRC |2-4x 0x4E|
;	+--------+----+----+----------------------------+-----+---------+
;
;	First we need to find the correct sector using the "readsection"
;	routine. When we have found the correct ID record we need to 
;	start writing after the GAP that follows the ID record. Thats
;	why the "readsection" routine always initiates a timer when it
;	has itself sync'd to the preamble and the sync marks. We need to
;	start writing when the timer overflows. 
;	In fact the write splice needs not be written bei the sector write
;	routine
;
	.macro	w6cycles
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	.endmacro
	
	.macro	wcycles
	.set cycles = @0
	.if ((cycles < 8) || (cycles > 1000))
	.error "Count must be between 8 and 1000)"
	.endif
	push	count
	ldi		count, (cycles/4)-1
l1:
	nop
	dec		count
	brne	l1
	pop		count
	.if (cycles & 1)
	nop
	.endif
	.if (cycles & 2)
	rjmp	PC+1
	.endif
	.endmacro
;
;	xxxx<yyyy^W		Write Buffer at yyyy to Sector xxxx
;
;		a4	= Sector Address
;		a1	= Buffer Address with data
;
;	Note that the MFMReader uses all the memory from 0x1000 to 0x3fff
;	but there is still enough room below 0x1000 to place the buffer
;	we want to write. So if you want to crate patterns and data to
;	be written use a buffer below 0x1000 (the MFMRead uses only very
;	little low memory that starts at 0x100)
;

writepattern:
	cbi		PORTD, MOT
	cbi		PORTD, DS
	ldi		count, 0
	rcall	delayms				; wait 256ms
	ldi		count, 0
	rcall	delayms				; wait 256ms
	lds		temp, a4l			; get sector requested
	tst		temp
	brne	writepattern005
	mov		temp, one
writepattern005:
	cpi		temp, numsect
	brlo	writepattern010
	ldi		temp, numsect
writepattern010:
	sts		writesector, temp
;
;	Prepare the sectorbuffer we want to write. The buffer includes
;	the Data Mark (0xFB), data, CRC and the GAP (22*0x4E). Also
;	generate the CRC.
;
	lds		xl, a1l
	lds		xh, a1h				; get memory buffer to write
	
	ldi		yl, low(sectorbuffer)
	ldi		yh, high(sectorbuffer)
	mov		crcl, ff			; Initiate CRC
	mov		crch, ff
	ldi		temp, 0xA1			; Note CRC calculation inlcudes
	updcrc	temp				; the sync marks and the data mark
	ldi		temp, 0xA1
	updcrc	temp
	ldi		temp, 0xA1
	updcrc	temp
	ldi		temp, 0xFB
	st		Y+, temp			; Save data mark in buffer
	updcrc	temp
	clr		count				; Works only for sector sizes of 512bytes
								; but without someone telling us we don't
								; know, we could read an arbitrary ID record
								; to retrieve the sector length but that could
								; affect performance, on the other hand we could
								; also remember the sector length form any ID
								; record and keep that as reference, it is very
								; unlikely that a floppy is formatted with
								; different sector sizes, although it could.
								; CRC generation on the fly is not working
								; the timing is too tight.

writepattern020:
;
;	Each Loop deals with 2 bytes
;
	ld		temp, X+			; copy memory buffer to sector buffer
	st		Y+, temp
	updcrc	temp				; and update CRC at the same time
	ld		temp, X+
	st		Y+, temp
	updcrc	temp
	dec		count
	brne	writepattern020
	

	st		Y+, crch			; Add CRC
	st		Y+, crcl
	ldi		temp, 0x4E			; Add GAP
	ldi		count, 48
writepattern040:
	st		Y+, temp
	dec		count
	brne	writepattern040

	rcall	write_nextidsegment
	sbi		PORTD, DS
	sbi		PORTD, MOT
	ret
	
write_nextidsegment:
	ldi		xl, low(7)				; Bytes to collect
	ldi		xh, high(7)
	ldi		yl, low(idsegment)
	ldi		yh, high(idsegment)
	clr		miss
	rcall	readsection
	brcc	write_readok
	ldi		char, '?'
	call	serout
	ret
	
;
;	And do now the "house-keeping"
;
write_readok:
	mov		crcl, ff
	mov		crch, ff
	ldi		temp, 0xA1
	updcrc	temp
	ldi		temp, 0xA1
	updcrc	temp
	ldi		temp, 0xA1
	updcrc	temp
	lds		temp, idsegment
	cpi		temp, 0xFE
	brne	write_nextidsegment		; It's not an idsegment
	updcrc	temp
	lds		temp, idsegment+1		; Track
	updcrc	temp
	lds		temp, idsegment+2		; Side
	updcrc	temp
	lds		temp, idsegment+3		; Sector
	updcrc	temp
	lds		temp, idsegment+4		; Length
	updcrc	temp
	lds		temp, idsegment+5
	cp		crch, temp
	lds		temp, idsegment+6
	cpc		crcl, temp
	breq	write_crcok
	ldi		char, '!'
	call	serout
	ret
write_crcok:
	lds		temp, idsegment+3
	lds		r0, writesector
	cp		temp, r0
	breq	write_waitgap
	rjmp	write_nextidsegment
	
;
;	The readsection always starts timer 1 when it has found a sync mark.
;	We only need to wait for the overflow flag to know when we need to
;	start writing the sector
;

write_waitgap:
write_waitgap010:
	sbis	TIFR1, TOV1
	rjmp	write_waitgap010
	cbi		PORTC, WG	
;
;	96 short intervals ......SSSSSSSSSSSSSS
;
	ldi		count, 12*8
wp0010:
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	19
	dec		count
	brne	wp0010
	nop
;
;	followed by three Sync Marks MLMLMSLMLMSLMLM
;
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	38				; M
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	54				; L
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	38				; M
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	54				; L
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	38				; M
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	22				; S
;
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	54				; L
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	38				; M
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	54				; L
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	38				; M
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	22				; S
;
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	54				; L
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	38				; M
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	54				; L
	cbi		PORTC, WD
	w6cycles
	sbi		PORTC, WD
	wcycles	38-6			; M-shortened to compensate initialisation
;
;	We use the following defintions
;
;	T-bit		Previous Clock Type '1'=data, '0'=clock
;	Char		The current byte
;	Count		Number of bits left in char
;	Y			Pointer to the buffer
;	X			Number of bytes to transmit (must be positive number)
;
;	Initiate the state
;
	set							; Sync Mark ends with a data pulse
	clr		count				; We have no bits in char
;	ldi		xl, low(1+512+2+38)
;	ldi		xh, high(1+512+2+38)
	ldi		xl, low(1+512+2)
	ldi		xh, high(1+512+2)
	ldi		yl, low(sectorbuffer)
	ldi		yh, high(sectorbuffer)

;	We are now here and need to create the pulse to finishe the
;	last M interval of the third SYNC MARK
;			------------->
;	Data	1 0 1 0 0 0 0 1| 1
;	Clock	 0 0 0 1 1 1 0 |0
;	MFM		100010010001001|01
;
;
;	Logic to generate MFM is very simple. We need only to remember
;	one state, that is if the previous pulse was a clock or a data
;	pulse.
;
;	get_next_bit()
;	if (previous_pulse == clock_pulse)
;	{
;		if (next_data_bit == '0')
;		{
;			prev next
;			---- ---- 
;	Data:	   0|   0|
;	Clock:	 1  | 1  |
;	MFM:	0100 0100 
;
;			interval=2usec
;			previous_pulse=clock_pulse
;		}
;		else
;		{
;			prev next
;			---- ---- 
;	Data:	   0|   1|
;	Clock:	 1  | 0  |
;	MFM:	0100 0001 
;
;			interval=3usec
;			previous_pulse=data_pulse
;		}
;	}
;	else
;	{
;		if (next_data_bit == '0')
;		{
;			get_next_bit()
;			if (next_data_bit == '0')
;			{
;			prev next
;			---- ---- ---- 
;	Data:	   1|   0|   0|
;	Clock:	 0  | 0  | 1  |
;	MFM:	0001 0000 0100 
;
;				interval=3usec
;				prevous_pulse=clock_pulse
;			}
;			else
;			{
;
;			prev next
;
;			---- ---- ---- 
;	Data:	   1|   0|   1|
;	Clock:	 0  | 0  | 0  |
;	MFM:	0001 0000 0001 
;
;			interval=4usec
;			prevous_pulse=data_pulse
;		}
;		else
;		{
;			prev next
;			---- ---- 
;	Data:	   1|   1|
;	Clock:	 0  | 0  |
;	MFM:	0001 0001 
;
;			interval=2usec
;			previous_pulse=data_pulse
;		}
;	}
;
;	We first generate a pulse using first a cbi some nop and a sbi
;	In case we have a clock of 16MHz and we want to create pulses
;	of 500ns this looks like the following.
;
w0000:
	cbi		PORTC, WD
	rjmp	PC+1				; rjmp PC+1 uses one instruction word
	rjmp	PC+1				; and 2 cycles, its 2 nops in one instruction
	rjmp	PC+1
	sbi		PORTC, WD
;
;	Each pulse is followed by at least 1500nsec silence. This corresponds
;	to 24 cycles. However the pulse generation already took 2 cycles so
;	we have 22 cycles left.
;									cycles
w0010:
	dec		count				;	1	More bits to go
	brmi	w0020				;	2	Need a byte

	nop							;		Compensate branch not taken
	rjmp	PC+1				;		Instead of getting a new byte
	rjmp	PC+1				;	 	we need to waste time
	rjmp	w0030				;
;
;	Getting a new byte
;	-	if we reached the end of the buffer then we are done
;	-	else set the number of bits we need to process to 7
;		because we are going to immediately process a bit
;	-	get the new byte
;
w0020:
	sbiw	X, 1				;	2	More bytes to go
	brmi	w0090				; 	1	No we are done
	ldi		count, 7			;	1	We will have 7 bits left
	ld		char, Y+			;	2	Get next char from buffer
w0030:
	lsl		char				;	1	Get next bit to the Carry bit
								;  ==
								;  10	cycles
;	We now have 4 possibilities
;	25-02-2019	start to share code so the branch to w0090 reaches
;	to the end of the block so we can immediately proceed with the
;	GAP
;
;	T=C=0
;	T=0, C=1
;	T=1, C=0
;	T=C=1
;
;	Some info about T-bit processing:
;	If interval is an even number of usec then T does not change if it is
;	odd it changes, in other words it changes whenever the interval is 3usec.
;
	brts	w0050				;		decision branch 1 or 2 cycle
	brcc	w0040				;		decision branch 1 or 2 cycle
;
;
;	TC=0, C=1
;
								;	2	cycles from two branches not taken
								;		and second branch taken
;
;	Required Interval 3usec				 48 cycles
;	Pulse creation						-10 cycles
;	Check for new character				-10	cycles
;	Decision Branches					- 2	cycles
;	Cycles left							 26	cycles
;
	set							;	1
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
								;  ==
								;  17	cycles and fall through
;
;	T=C=0
;
;
w0040:
								;	3	cycles from first branch not taken and second taken
;
;	Required Interval 2usec				 32 cycles
;	Pulse creation						-10 cycles
;	Check for new character				-10	cycles
;	Decision Branches					- 3	cycles
;	Cycles left							  9	cycles
;
;
	nop							;   1
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	w0000				;	2
								;  ==
								;   9	cycles
;
w0050:
	brcs	w0080				;
;
;	T=1, C=0
;
								;	3	cycles from first branch taken
								;		and second branch not taken
;
;	We need another bit
;
	dec		count				;	1	More bits to go
	brmi	w0060				;	2	Need a byte

	nop							;		compensate branch not taken
	rjmp	PC+1				;		Instead of getting a new byte
	rjmp	PC+1				;		we need to waste time
	rjmp	w0070


w0060:
	sbiw	X, 1				;	2	More bytes to go
	brmi	w0100				; 	1	No we are done
	ldi		count, 7			;	1	We will have 7 bits left
	ld		char, Y+			;	2	Get next char from buffer

w0070:
	lsl		char				;	1
								;  ==
								;  10	cycles
;
	brcc	w0075				;
;
;	Next bit is '1'
;
;	Required Interval 4usec				 64 cycles
;	Pulse creation						-10 cycles
;	Check for new character				-10	cycles
;	Decision Branches					- 3	cycles
;	Check for new character				-10 cycles
;	Decision Branch						- 1	cycle
;	Cycles left							 30	cycles
;
;	
;
;
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	w0080				;	2	
								;  ==
								;  22	and fall through
;
;	Next bit is '0'
;
;	Required Interval 3usec				 48 cycles
;	Pulse creation						-10 cycles
;	Check for new character				-10	cycles
;	Decision Branches					- 3	cycles
;	Check for new character				-10 cycles
;	Decision Branch						- 2	cycle
;	Cycles left							 13	cycles
;
w0075:
	clt							;	1
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
								;  ==
								;   5	and fall through
;
w0080:
;
;	T=C=1
;
								;	4	cycles from both branch taken
;
;	Required Interval 2usec				 32 cycles
;	Pulse creation						-10 cycles
;	Check for new character				-10	cycles
;	Decision Branches					- 4	cycles
;	Cycles left							  8	cycles
;
;	So far we have spent 14 cycles (10 cycles from the buffer check and 2 x 2
;	cycles from branches taken). Previous pulse was a data pulse and next
;	bit is '1' so the interval is usec, so we need to waste 8 cycles
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	PC+1				;	2
	rjmp	w0000				;	2
								;  ==
								;   8

;---------------------------------------------------------
;	
;	All data bits written, add gap
;
;	a)	when we arrive from label w0020, if the gap byte
;		would be part of the buffer the lsl at w0030 would
;		set C=0
;
;
w0090:
	brtc	w0091
;
;	T=1, C=0 this case requires another bit and we know
;	this bit is the second bit of the gap byte 0x4E, i.e.
;	this bit is '1'
;
;	Required Interval 4usec				 64 cycles
;	Pulse creation						-10 cycles
;	Check for last character			- 7	cycles
;	Branch not taken					- 1 cycles
;	Cycles left							 46	cycles

	rjmp	w0105				;  2


w0091:
;
;	T=0, C=0 
;
;
;	Required Interval 2usec				 32 cycles
;	Pulse creation						-10 cycles
;	Check for last character			- 7	cycles
;	Branch taken						- 2	cycles
;	Cycles left							 13	cycles
;
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	nop
	cbi		PORTC, WD
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	sbi		PORTC, WD
;
;	Then we need to process the next bit
;	T=0, C=1
;
;	Required Interval 3usec			 48	cycles
;	Pulse creation					-10 cycles
;	Cycles left						 38 cycles
;
	rjmp	w0110				;  2


;
;	b)	when we arrive from label w0060, if the gap byte
;		would be part of the buffer the lsl at w0070 would
;		set C=0, which is the same as to proceed to w0075
w0100:
;	
;	Required Interval 3usec				 48 cycles
;	Pulse creation						-10 cycles
;	Check for new character				-10	cycles
;	Decision Branches					- 3	cycles
;	Check for last character			- 7 cycles
;	Cycles left							 18	cycles
;
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	cbi		PORTC, WD
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	sbi		PORTC, WD

;
;	Then we need to process the next bit 
;	T=0 (as would be the case at w0075), C=1
;
;	Required Interval 3usec			 48	cycles
;	Pulse creation					-10 cycles
;	Cycles left						 38 cycles
;
	rjmp	w0110				;  2

w0105:
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1

w0110:
	wcycles	36
	cbi		PORTC, WD
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	sbi		PORTC, WD
	nop
	nop
	nop
	nop
	sbi		PORTC, WG
	ret

;
;	Some thoughts about a format routine. A track holds
;	12500 bytes of raw data. From the IBM Format the track
;	is formatted as follows
;
;	80 x 0x4E
;	12 x 0x00
;	 3 x 0xC2		; Index Sync Mark
;	 1 x 0xFC
;	50 x 0x4E
;	each sector
;	12 x 0x00
;	 3 x 0xA1		; Record Sync Mark
;	 1 x 0xFE		; Address Mark
;	 1 x track
;	 1 x head
;	 1 x sector
;	 1 x length
; 	 2 x CRC
;	22 x 0x4E
;	12 x 0x12
;	 3 x 0xA1
;	 1 x 0xFB		; Data Mark
;	nn x data
;	 2 x CRC
;	84 x 0x4E
;	
;	
;
;

;
;			-->
;	Data	0 1 0 0 1 1 1 0| 0 1 0 0 1 1 1 0|
;	Clock	 0 0 1 0 0 0 0 |1 0 0 1 0 0 0 0 |
;	MFM		001001001010100|1001001001010100|
;
;	After the data we would write the following pattern as gap
;
;	MMSSMMMMSSMMMMSSM (the pattern MMMSSM repeated 22 times)
;
;	followed by
;
;	SSSSSSSS (12 x 8 = 96 short pulses)
;
;	and a SYNC MARK
;
;	LMLMSLMLMSLMLM
;
;



;--------------------------------------------------------------------------
;
;	write sector with support of the CPLD 
;
;	xxxx<yyyy^U		Write Buffer at yyyy to Sector xxxx using CPLD
;
;		a4	= Sector Address
;		a1	= Buffer Address with data
;
;	Note that the MFMReader uses all the memory from 0x1000 to 0x3fff
;	but there is still enough room below 0x1000 to place the buffer
;	we want to write. So if you want to create patterns and data to
;	be written use a buffer below 0x1000 (the MFMReader uses only very
;	little low memory that starts at 0x100)
;

writecpld:
	sts		sectmo, zero		; Only try 256 times to read a ID record
;	ldi		r24, low(seclist)	; Debug
;	ldi		r25, high(seclist)
	cbi		PORTD, MOT
	cbi		PORTD, DS
	ldi		count, 0
	rcall	delayms				; wait 256ms
	ldi		count, 0
	rcall	delayms				; wait 256ms
	lds		temp, a4l			; get sector requested
	tst		temp
	brne	writecpld005
	mov		temp, one
writecpld005:
	cpi		temp, numsect
	brlo	writecpld010
	ldi		temp, numsect
writecpld010:
	sts		writesector, temp
;
;	Prepare the sectorbuffer we want to write. The buffer includes
;	the Data Mark (0xFB), data, CRC and the GAP (22*0x4E). Also
;	generate the CRC.
;
	lds		xl, a1l
	lds		xh, a1h				; get memory buffer to write
	
	ldi		yl, low(sectorbuffer)
	ldi		yh, high(sectorbuffer)
	mov		crcl, ff			; Initiate CRC
	mov		crch, ff
	ldi		temp, 0xA1			; Note CRC calculation inlcudes
	updcrc	temp				; the sync marks and the data mark
	ldi		temp, 0xA1
	updcrc	temp
	ldi		temp, 0xA1
	updcrc	temp
	ldi		temp, 0xFB
	st		Y+, temp			; Save data mark in buffer
	updcrc	temp
	clr		count
writecpld020:
;
;	Each Loop deals with 2 bytes
;
	ld		temp, X+			; copy memory buffer to sector buffer
	st		Y+, temp
	updcrc	temp				; and update CRC at the same time
	ld		temp, X+
	st		Y+, temp
	updcrc	temp
	dec		count
	brne	writecpld020

	st		Y+, crch			; Add CRC
	st		Y+, crcl
	ldi		temp, 0x4E			; Add GAP
	ldi		count, 8
writecpld040:
	st		Y+, temp
	dec		count
	brne	writecpld040

	rcall	writecpldreadsegment
	ldi		count, 0
	rcall	delayms				; wait 256ms
	sbi		PORTD, DS
	sbi		PORTD, MOT
	ret
	
writecpldreadsegment:
	ldi		xl, low(7)				; Bytes to collect
	ldi		xh, high(7)
	ldi		yl, low(idsegment)
	ldi		yh, high(idsegment)
	clr		miss
	rcall	quicksection			; Read 8 bytes of the next record
	brcc	writecpldreadok
	ldi		char, '?'
	call	serout
	ret
	
;
;	And do now the "house-keeping", e.g. check the record just
;	read. It should be an address record with correct CRC and
;	then we check if it is the desired sector to write
;
writecpldreadok:
;
;	Debug
;
;	movw	xh:xl, r25:r24
;	lds		temp, idsegment+0
;	st		X+, temp
;	lds		temp, idsegment+1
;	st		X+, temp
;	lds		temp, idsegment+2
;	st		X+, temp
;	lds		temp, idsegment+3
;	st		X+, temp
;	movw	r25:r24, xh:xl

	lds		temp, sectmo
	dec		temp
	sts		sectmo, temp
	brne	writecpldnotmo
	rjmp	writecpldtmo
;
;
;
writecpldnotmo:


	mov		crcl, ff
	mov		crch, ff
	ldi		temp, 0xA1
	updcrc	temp
	ldi		temp, 0xA1
	updcrc	temp
	ldi		temp, 0xA1
	updcrc	temp
	lds		temp, idsegment
	cpi		temp, 0xFE
	brne	writecpldreadsegment	; It's not an idsegment
	updcrc	temp
	lds		temp, idsegment+1		; Track
	updcrc	temp
	lds		temp, idsegment+2		; Side
	updcrc	temp
	lds		temp, idsegment+3		; Sector
	updcrc	temp
	lds		temp, idsegment+4		; Length
	updcrc	temp
	lds		temp, idsegment+5
	cp		crch, temp
	lds		temp, idsegment+6
	cpc		crcl, temp
	breq	writecpldsectorok
	ldi		char, '!'				; CRC read error
	call	serout
	ret
writecpldsectorok:
	lds		temp, idsegment+3
	lds		r0, writesector
	cp		temp, r0
	breq	writecpldbuffer			; Yes our sector
	rjmp	writecpldreadsegment


writecpldtmo:
	call	print
	.db		"Sector not found.", CR, LF, 0
	ret
;
;	The readsection always starts timer 1 when it has found a sync mark.
;	We only need to wait for the overflow flag to know when we need to
;	start writing the sector
;

writecpldbuffer:
	
	
	cbi		PORTB, READ				; Disable CPLD Data Outputs 
	out		DDRA, ff				; Data Port 
	out		PORTA, zero
;
;	DATA8 normally used as DRQ when reading is used as suppress clock
;	pulse for sync marks when in write mode. 
;

	sts		TCCR2B, zero			; Stop CPLD clock
	sts		TCCR2A, zero
	sts		TCNT2, zero
	ldi		temp, 15				; 1=8MHz, 3=4MHz, 7=2MHz, 15=1MHz
	sts		OCR2A, temp
	lsr		temp
	sts		OCR2B, temp
	ldi		temp, (0<<COM2A1) | (0<<COM2A0) | (1<<COM2B1) | (1<<COM2B0) | (1<<WGM21) | (1<<WGM20)
	sts		TCCR2A, temp
	ldi		temp, (1<<WGM22) | (0<<CS22) | (0<<CS21) | (1<<CS20)
	sts		TCCR2B, temp			; Start CPLD clock
;
;
;
.equ	sectorsizecpld	= 1+512+2+2 ; 1 x header, 512 x data, 2 x crc, 2 x writesplice
	ldi		xl, low(sectorsizecpld)	
	ldi		xh, high(sectorsizecpld)
	ldi		yl, low(sectorbuffer)
	ldi		yh, high(sectorbuffer)
	out		PORTA, zero				; Make sure the next byte written is 0x00
	sbi		PORTB, 0
	cbi		PORTB, 0
writecpldbuffer010:					; wait for sync timer 1 to know when we can start
	sbis	TIFR1, TOV1				; writing
	rjmp	writecpldbuffer010

	sbi		PORTB, ACK
	cbi		PORTB, ACK
	ldi		count, 12				; 12 x 0x00 preamble
	cbi		PORTC, WG				; enable write gate
;
;	The new interface with DRQ for writes as well. It's much faster to have
;	a request/acknowledge interface than detecting an edge. Here we have combined
;	the acknowledgement with the loading of a new byte. So we wait for DRQ and
;	then load a new byte which at the same time clears DRQ. The loop takes 12
;	cycles plus eventually 3 cycles if we have to wait. The total time between
;	two bytes is 16sec (1.6usec for hard disks), in other words about 25 cycles
;	when running at 16MHz. For a hard disk interface we plan to run the MCU with
;	it's maximum official speed, 20MHz, so 1.6usec corresponds to 32 cycles. In
;	other words we have plenty of time.
;
writecpldpre:
	sbis	PINB, DRQ				;	 2	Wait for DRQ 2 cycles if DRQ is set
	rjmp	PC-1					;	 3	If sbis not take a total of 3 cycles
	cbi		PORTD, A1SYNC			;	 2
	out		PORTA, zero				;	 1
	sbi		PORTB, ACK				;	 2
	cbi		PORTB, ACK				;	 2
	dec		count					;	 1
	brne	writecpldpre			;	 2
									;	==
									;	12+n*3 
;
	ldi		char, 0xA1				;	 1
	ldi		count, 3				;	 1	3 x 0xA1 sync marks
writecpldsync:
	sbis	PINB, DRQ				;		Wait for DRQ
	rjmp	PC-1
	sbi		PORTD, A1SYNC
	out		PORTA, char
	sbi		PORTB, ACK
	cbi		PORTB, ACK
	dec		count
	brne	writecpldsync
;
;	The real data cycles take a little bit longer, but nothing to be worried about.
;	We are well below the 25cycles. 
;
writecpldbuffer050:					; Data buffer with Data Mark and CRC and write splice
	ld		temp, Y+				;	 2
	sbis	PINB, DRQ				;	 2	Wait for DRQ
	rjmp	PC-1					;	(3)
	cbi		PORTD, A1SYNC			; 	 2
	out		PORTA, temp				; 	 1
	sbi		PORTB, ACK				;	 2
	cbi		PORTB, ACK				;	 2
	sbiw	X, 1					;	 2
	brpl	writecpldbuffer050		;	 2	another byte
									;	==
									;	15+n*3 
;
;	Wait until the byte has been written
;
;	sbis	PINB, DRQ
;	rjmp	PC-1
;	
	sbi		PORTC, WG				; disable write gate
	out		DDRA, zero
	sbi		PORTB, READ
	sts		TCCR2B, zero			; Stop CPLD clock
	sts		TCCR2A, zero			; Stop CPLD clock
	sts		TCNT2, zero
	ldi		temp, 0x01				; 1=8MHz, 3=4MHz
	sts		OCR2A, temp
	lsr		temp
	sts		OCR2B, temp
	ldi		temp, (0<<COM2A1) | (0<<COM2A0) | (1<<COM2B1) | (1<<COM2B0) | (1<<WGM21) | (1<<WGM20)
	sts		TCCR2A, temp
	ldi		temp, (1<<WGM22) | (0<<CS22) | (0<<CS21) | (1<<CS20)
	sts		TCCR2B, temp			; Start CPLD clock
	ret
	
	
#ifdef mfmenctimer
;--------------------------------------------------------------------------
;
;	Using Timer 2 to generate write pulses
;
	out		TCCR2B, zero	; stop timer
	sbi		TIFR2, OCF2A	; reset overflow flag
	ldi		temp, (1<<COM2B1) | (1<<COM2B0) | (1<<WGM21)
	out		TCCR2A, temp

	ldi		temp, 8		;
	out		OCR2B, temp	; write pulse is low for 500ns
	
	ldi		temp, 32
	mov		ishort, temp
	ldi		temp, 48
	mov		imedium, temp
	ldi		temp, 64
	mov		ilong, temp

	sbi
	out		TCNT2, zero	; reset timer 
	out		OCRA, ishort	; short interval
	out		TCCR2B, one	; start timer


;--------------------------------------------------------------------------
;
;	Timer 2 is in CTC mode. OCR2A defines the period. We use OC2B 
;	to define the pulse. In CTC mode OC2B is cleared at top and 
;	set at OCR2B. OCR2B is set ot 8 which gives a 500ns write pulse.
;
;	To encode MFM we continuously adjust OCR2A, however we need to
;	make sure that we change it only the previous interval has
;	finished
;
	.macro	nextpulse
	sbis	TIFR2, OCF2A
	rjmp	PC-1
	sbi	TIFR2, OCF2A
	out	OCR2A, @0
	.endmacro
;--------------------------------------------------------------------------
;
;	State 00
;
;	The two previous data bits were 00. This implies that we had a
;	clock pulse
;		----->
;	Data	0   0|
;	Clock	  1  |
;	MFM	0 1 0|
;
state00:
	dec		count
	brpl		l0010
	sbiw		xh:xl, 1
	breq		s0000
	ldi		count, 7
	ld		char, Y+
l0010:
	lsl		char
	brcs		l0015
;
;	If the next bit is 0 then we need to create a another clock
;	pulse, or in other words a short interval 
;
;		----->
;	Data	0   0|  0
;	Clock	  1  |1 
;	MFM	0 1 0|1 0
;
	nextpulse	ishort
	rjmp		state00
;
;	If the next bit is 1 then we need to create a data pulse, or
;	in other words a medium interval  
;
l0015:
	nextpulse	imedium
;
;	Fall through to state 01
;--------------------------------------------------------------------------
;
;	State x1
;
;	The previous data bit was 1. This implies that we had a
;	data pulse
;		----->
;	Data	x   1|
;	Clock	  0  |
;	MFM	x 0 1|
;
statex1:
	dec		count
	brpl		l0020
	sbiw		xh:xl, 1
	breq		s0000
	ldi		count, 7
	ld		char, Y+
l0020:
	lsl		char
	brcc		state10
;
;	if the next bit is 1 then we need to create another data pulse,
;	or in other words a short interval 
;
;		----->
;	Data	x   1|  1
;	Clock	  0  |0 
;	MFM	x 0 1|0 1
;
	nextpulse	ishort
	rjmp		statex1
;--------------------------------------------------------------------------
;
;	State 10
;
;	The two previous data bits were 10. This implies that we had a
;	data pulse but we need another bit to decide the next interval
;		----->
;	Data	1   0|
;	Clock	  0  |
;	MFM	1 0 0|
;
state10:
	dec		count
	brpl		l0030
	sbiw		xh:xl, 1
	breq		s0000
	ldi		count, 7
	ld		char, Y+
l0030:
	lsl		char
	brcc		l0040
;
;	if the next bit is 1 then we need to create a data pulse,
;	or in other words a long interval 
;
;		----->
;	Data	1   0|  1
;	Clock	  0  |0 
;	MFM	1 0 0|0 1
;
	nextpulse	ilong
	rjmp		statex1
;
;	if the next bit is a 0 then we need to create a clock pulse,
;	or in other words a medium intervals
;		----->
;	Data	1   0|  0
;	Clock	  0  |1 
;	MFM	1 0 0|1 0
;
l0040:
	nextpulse	imedium
	rjmp		state00
;
;
;
s0000:
	out		TCCR2A, zero	
	
	

























#endif
