;--------------------------------------------------------------------------
;
;	MFM Floppy Disk Reader
;
;	Proof of concept to read HD Floppy without FDC. The basic idea is to
;	use a timer that measures the time between pulses of the RD signal.
;	Together with a statemachine we reconstruct the bytes. First we will
;	read and write to a floppy drive and later we attempt to emulate a
;	floppy drive;
;
;	2019-02-23	Notes for a Floppy Emulator
;
;	A floppy emulator would have to continuously send the data from a track
;	in other words it will continuously send the MFM data stream of the 
;	selected track.
;
;	We could actually think about sending the data at our pace. That is
;	read the block from the SD-Card that corresponds to the next sector
;	send the ID and then the Data section. Reading a block from the
;	SD-Card takes about 2ms. The floppy controller will not time out
;	as it typically waits several (4) index pulses before giving up.
;	This would add about 18*2ms to the track. As the floppy controller
;	has no measure of rotational speed it is save to say that this
;	does not affect operation. We could indeed repeat the following
;
;	-	create index pulse
;	-	repeat 18 times
;		-	read SD-Card
;		-	send ID record
;		-	send Data record
;	-	send GAP
;
;	We always keep the current sector being sent. When the host wants
;	to write a sector it will send the data section right after the
;	ID section. We could actually connect WG to an interrupt line
;	and just receive the data section, extract the data and write it
;	to the SD-Card. We could even be more daring. After the ID section
;	has been sent we send 22 x 0x4E and 12 x 0x00. We could check WG
;	after the last 0x4E for every 0x00. The host first sends 12 x 0x00
;	which we can savely ignore and just activate sync recognition.
;
;	We also need to send index pulses
;
;	2019-03-02	Made backup (extension -20190302) for both folders,the
;	AVR and the WinCUPL project.
;
;	2019-03-02	New MCU/CPU Interface
;
;	CLK		the variable clock
;	DRQ		this now reflects the data request for read and writes. 
;	READ	this signal replaces OE, logic is now positive and controls
;			the output enable of the data port on the CPLD, the source
;			of the DRQ and controls the DATA register
;	ACK		either acknowledges thet we retrieved a byte or writes the
;			byte
;	A1SYNC	set if a sync mark needs to be written
;	RES		resets the MFM decoder statemachine
;
;
.include	"m1284Pdef.inc"
.listmac
.include	"../include/macro-library-v1-0.asm"

.def	miss	= r8				; Missed Interval Counter
.def	ff		= r9				; Constant 0xFF
.def	one		= r10				; Constant 0x01
.def	zero 	= r11				; Constant 0x00
.def	crcl	= r12				; Used to calculate CRC
.def	crch	= r13				;            "
.def	crc3	= r14
.def	crc4	= r15
.def	temp 	= r16				; Temporay register with immediate always destroyed!
.def	char	= r17				; The character or byte
.def	count	= r18				; General purpose counter
;
;	Definitions for the interactive monitor
;
#define clibuffer InputBuffer		; use this buffer for monitor
#define charout   serout
#define charin    serin


.equ	F_CPU		= 22118400		; Crystal clock
;.equ	F_CPU		= 20000000		; Crystal clock
;.equ	F_CPU		= 18432000		; Crystal clock
;.equ	F_CPU		= 16000000		; Crystal clock

;.equ	BAUD0		= 38400			; Baudrate USART0
.equ	BAUD0		= 115200		; Baudrate USART0
.equ	BAUD1		= 9600			; Baudrate USART1

.equ	CR		=	0x0d		; carriage return
.equ	LF		=	0x0a		; line feed
.equ	ESC		=	0x1b		; ESC to exit
;
;	Read Pulses procuded by MFM occur with an interval of 2usec,
;	3usec or 4usec for HD 3 1/2 disks. For DD it would be 4,6 and
;	8usec respectively. When we measure the interval timer we use
;	the 8 bit counter 0 in free running mode. At each falling
;	edge of the RD signal we read the timer and set the counter
;	to zero. The counter just read tells us how many clock
;	cycles have been passed between the RD pulses. We compare
;	them with the value that corresponds to 2.5usec (t_short)
;	and 3.5usec (t_long), everthing lower than t_short is
;	considered to be a 2usec interval, everything higher than
;	t_long is consered to be a 4usec interval and everything
;	in between is considered to be a 3usec interval.
;	
;	The values depend on the data rate of the floppy format
;	and the MCU clock. Here we just assume HD diskettes
;
.equ	t_long	= int(3.5*F_CPU/1000000)
.equ	t_short	= int(2.5*F_CPU/1000000)	
.equ	numsect	= 18

;--------------------------------------------------------------------------
;
; Baud Rate Calculator
;
.equ	UBRR0_VAL   = ((F_CPU+BAUD0*8)/(BAUD0*16)-1)	; nearest value
.equ	BAUD0_REAL  = (F_CPU/(16*(UBRR0_VAL+1)))		; effective baudrate
.equ	BAUD0_ERROR = ((BAUD0_REAL*1000)/BAUD0-1000)	; error in 0/00
.if ((BAUD0_ERROR>10) || (BAUD0_ERROR<-10))
	.error " USART0: Effective baudrate is more then 1 percent off requested baudrate"
.endif
.equ	UBRR1_VAL   = ((F_CPU+BAUD1*8)/(BAUD1*16)-1)	; nearest value
.equ	BAUD1_REAL  = (F_CPU/(16*(UBRR1_VAL+1)))		; effective baudrate
.equ	BAUD1_ERROR = ((BAUD1_REAL*1000)/BAUD1-1000)	; error in 0/00
.if ((BAUD1_ERROR>10) || (BAUD1_ERROR<-10))
	.error " USART1: Effective baudrate is more then 1 percent off requested baudrate"
.endif



;--------------------------------------------------------------------------
;
;	Data
;
	.dseg
track:			.byte	1
seccnt:			.byte	1
missed:			.byte	1
writesector:	.byte	1
sectmo:			.byte	1
idsegment:		.byte	8
InputBuffer:	.byte	80			; for the "Apple II" monitor
;	align	8
;seclist:		.byte	256			; Debug buffer 
;--------------------------------------------------------------------------
;
;	Flash
;
	.cseg
.org	0
	jmp		reset
;.org	INT0addr
;	rjmp	INT0_isr
;.org	INT1addr
;	rjmp	INT1_isr
;.org	INT2addr
;	rjmp	INT2_isr
;.org	PCI2addr
;	rjmp	PC_INIT
;.org	WDTaddr
;	rjmp	wdt_isr
;.org	OC0Aaddr
;	rjmp	OC0_isr
;.org	URXC0addr
;	rjmp	rxc0_isr
;.org	UDRE0addr
;	rjmp	dre0_isr
	
.org	INT_VECTORS_SIZE

	.dw	t_long						; Just as reference in the listing
	.dw	t_short


;--------------------------------------------------------------------------
;
;	Power Up and Reset
;
reset:
;
;	Set Stack Pointer to top of RAM
;
	ldi		temp,low(RAMEND) 
	out		SPL,temp 
	ldi		temp,high(RAMEND) 
	out		SPH,temp
;
;	Init Constants
;
	clr		zero
	clr		one
	inc		one
	clr		ff
	dec		ff

;--------------------------------------------------------------------------
;
;	USART0
;	
;	Interactive user interface to control the AVR
;
;	8 data, no parity, 1 stop bit, baud-rate see above
;
	ldi		temp, HIGH(UBRR0_VAL)
	sts		UBRR0H, temp
	ldi		temp, LOW(UBRR0_VAL)
	sts		UBRR0L, temp
;
;	Enable receiver and transmitter 
;
	ldi		temp, (1<<RXEN0)|(1<<TXEN0)
	sts		UCSR0B, temp
;
;	Set frame format: 8 data, 1 stop, no parity bit
;
	ldi		temp, (3<<UCSZ00)
	sts		UCSR0C, temp
;
;			
;							 __   __
;	(PCINT8/XCK0/T0)	PB0	[  \_/  ] PA0	(ADC0/PCINT0)
;	(PCINT9/CLKO/T1)	PB1	[       ] PA1	(ADC1/PCINT1)
;	(PCINT10/INT2/AIN0)	PB2	[       ] PA2	(ADC2/PCINT2)
;	(PCINT11/OC0A/AIN1)	PB3	[       ] PA3	(ADC3/PCINT3)
;	(PCINT12/OC0B/SS)	PB4	[       ] PA4	(ADC4/PCINT4)
;	(PCINT13/ICP3/MOSI)	PB5	[       ] PA5	(ADC5/PCINT5)
;	(PCINT14/OC3A/MISO)	PB6	[       ] PA6	(ADC6/PCINT6)
;	(PCINT15/OC3B/SCK)	PB7	[       ] PA7	(ADC7/PCINT7)
;				 /RESET	[       ] AREF
;				    VCC	[       ] GND
;				    GND	[       ] AVCC
;				  XTAL1	[       ] PC7	(TOSC2/PCINT23)
;				  XTAL2	[       ] PC6	(TOSC1/PCINT22)
;	(PCINT24/RXD0/T3)	PD0	[       ] PC5	(TDI/PCINT21)
;	(PCINT25/TXD0)		PD1	[       ] PC4	(TDO/PCINT20)
;	(PCINT26/RXD1/INT0)	PD2	[       ] PC3	(TMS/PCINT19)
;	(PCINT27/TXD1/INT1)	PD3	[       ] PC2	(TCK/PCINT18)
;	(PCINT28/XCK1/OC1B)	PD4	[       ] PC1	(SDA/PCINT17)
;	(PCINT29/OC1A)		PD5	[       ] PC0	(SCL/PCINT16)
;	(PCINT30/OC2B/ICP)	PD6	[_______] PD7	(OC2A/PCINT31)
;
;--------------------------------------------------------------------------
;
;		Port A - Data 
;
;	Port A is the byte wide interface to the CPLD to read and write
;	bytes. The CPLD does all the "MFM handling".
;
	out		DDRA, zero				; Initially configured as input

;--------------------------------------------------------------------------
;
;		Port B
;
;
;	During reads DRQ is an input and signals the MCU that a byte is 
;	waiting from the MFM read state machine. During writes this signal
;	is an output and tells the CPLD to write a SYNC Mark must be set
;	at the same time a new value is writen to Port A during write
;
.equ	DBG		=	0		; Debug signal for the LA		Output
.equ	READ	=	1		; CPLD Data Bus Output Enable	Output
.equ	ACK		=	2		; Acknowledge byte				Output
.equ	RES		=	3		; Reset mfmread state machine	Output
.equ	DRQ		=	4		; Data Request CPLD->MCU		Input

	ldi		temp, (1<<READ)
	out		PORTB, temp
	ldi		temp, (1<<DBG) | (1<<READ) | (1<<ACK) | (1<<RES)
	out		DDRB, temp
;--------------------------------------------------------------------------
;
;		Port C
;
;	Connects to various Floppy Drive Signals
;
.equ	DC		=	7		; Disk Change		Input
.equ	HS		=	6		; Head Select		Output
.equ	RD		=	5		; Read Data			Input
.equ	WP		=	4		; Write Protect		Input
.equ	TRK0	=	3		; Track 0			Input
.equ	WG		=	2		; Write Gate		Output
.equ	WD		=	1		; Write Data		Output
.equ	STEP	=	0		; Step				Output

	out		PORTC, ff
	ldi		temp, (1<<HS) | (1<<WG) | (1<<WD) | (1<<STEP)
	out		DDRC, temp

;--------------------------------------------------------------------------
;
;		Port D
;
;	Connects to various Floppy Drive Signals and CPLD
;
;	The base clock to the CPLD is variable. To read HD floppy it is set
;	to 8MHz, to write HD floppy it is set to 4MHz and for DD it is half
;	the rate. 
;	WEN is an output from the CPLD from the MFM Write state machine. A 
;	falling edge signals the MCU that it should place the next byte to
;	Port A. After the falling edge the MCU has approx 10usec to place
;	a new byte before the statemachine latches the value into the MFM
;	write shift register. For a AVR at 16MHz this is plenty of time.
;
.equ	DIR		=	7		; Direction			Output
.equ	CLK		=	6		; CPLD Clock		Output
.equ	DS		=	5		; Drive Select		Output
.equ	IDX		=	4		; Index Pulse		Input
.equ	MOT		=	3		; Motor On			Output
.equ	A1SYNC	=	2		; Write Enabled		Input

	ldi		temp, (1<<DIR) | (1<<CLK) | (1<<DS) | (1<<MOT) | (0<<A1SYNC) | (1<<IDX)
	out		PORTD, temp
	ldi		temp, (1<<DIR) | (1<<CLK) | (1<<DS) | (1<<MOT) | (1<<A1SYNC)
	out		DDRD, temp
;--------------------------------------------------------------------------
;
;		Timer 2
;
;	Fast PWM Mode 7
;	Output B Set on compare match, clear at bottom
;

	ldi		temp, 0x01			; 1=8MHz, 3=4MHz
	sts		OCR2A, temp
	lsr		temp
	sts		OCR2B, temp
	ldi		temp, (0<<COM2A1) | (0<<COM2A0) | (1<<COM2B1) | (1<<COM2B0) | (1<<WGM21) | (1<<WGM20)
	sts		TCCR2A, temp
	ldi		temp, (1<<WGM22) | (0<<CS22) | (0<<CS21) | (1<<CS20)
	sts		TCCR2B, temp
;--------------------------------------------------------------------------
;
;		Done
;
	rcall	splashscreen
	sts		a1l, zero
;	rcall	seek				; Do a home
	rjmp	mon


;--------------------------------------------------------------------------
;
;		Serial 
;
serout:
	push	char
seroutwait:
	lds		char, UCSR0A
	sbrs	char, UDRE0
	rjmp	seroutwait
	pop		char
	sts		UDR0,char
	ret			
serin:
	wdr
	lds		char, UCSR0A
	sbrs	char, RXC0
	rjmp	serin
	lds		char, UDR0
	ret

;--------------------------------------------------------------------------
;
;	Splashscreen
;
splashscreen:
	ldi		zl, low(2*hello)
	ldi		zh, high(2*hello)

splashsend:
	lpm		char ,Z+ 
	tst		char
	breq	splashdone
	rcall	serout
	rjmp	splashsend

splashdone:
	ret

;--------------------------------------------------------------------------
;
;	AVR Monitor, inlcude the default commands followed by application
;	specific commands
;
.include	"../include/monitor-chartbl-v1-3.asm"
	.db		0, '?'
	.db		0, 'O'			; 		Toggle Motor On		
	.db		0, 'S'			;		Toggle Drive Select
	.db		0, 'H'			;		Toggle Head Select
	.db		0, 'T'			;		Seek
	.db		0, 'R'			;		Read Track
	.db		0, 'K'			;		Analyze Track and show info
	.db		0, 'W'			;		Generate Write Pattern
	.db		0, 'Q'			;		Read Track using CPLD
	.db		0, 'I'			;
	.db		0, 'I' & 0x1F	;		
	.db		0, 'U'			;		Write using CPLD
	.db		0, 'J'			;		Print IDs
	.db		0, 'N'			;		Make CRC Table
	.db		0, 'Y'			;		CRC calculate aaaa.bbbby
	.db		0, 'Z'			;		CRC calculate optimized
	
;
;	AVR Monitor, inlcude the action routine address table of the default
;	commands followed by the addresses of the application specific 
;	command routines. Both tables must be in sync of course...
;
.include	"../include/monitor-subtbl-v1-3.asm"
	.dw		helpscreen
	.dw		togglemot
	.dw		toggleds
	.dw		togglehs
	.dw		seek
	.dw		readtrack
	.dw		analyzetrack
	.dw		writepattern
	.dw		quickread
	.dw		indexx
	.dw		initmfm
	.dw		writecpld
	.dw		printids
	.dw		crcmaketable
	.dw		crcupdate
	.dw		crcupdate2
	
.include	"MFMCRC32.asm"


initmfm:
	cbi		PORTB, RES
	sbi		PORTB, RES
	ret	
togglemot:
	ldi		temp, (1<<MOT)
	out		PIND, temp
	ret

toggleds:
	ldi		temp, (1<<DS)
	out		PIND, temp
	ret

togglehs:
	ldi		temp, (1<<HS)
	out		PINC, temp
	ret

printids:
	cbi		PORTD, MOT
	cbi		PORTD, DS
	ldi		count, 0
	rcall	delayms				; wait 256ms
	ldi		count, 0
	rcall	delayms				; wait 256ms
	ldi		temp, 18
	sts		seccnt, temp

printids010:

	lds		temp, a1l
	sbrs	temp, 1
	rjmp	printids020
	ldi		xl, low(8)
	ldi		xh, high(8)
	ldi		yl, low(pprint)
	ldi		yh, high(pprint)
	lds		temp, a1l
	sbrc	temp, 0
	rcall	readsection			; 3j
	lds		temp, a1l
	sbrs	temp, 0
	rcall	quicksection		; 2j

printids020:
	ldi		xl, low(8)
	ldi		xh, high(8)
	ldi		yl, low(pprint)
	ldi		yh, high(pprint)
	lds		temp, a1l
	sbrc	temp, 0
	rcall	readsection			; 1j
	lds		temp, a1l
	sbrs	temp, 0
	rcall	quicksection		; 0j


	call	print
	.db		"Section: 0x", 0x80, " 0x", 0x81, " 0x", 0x82, " 0x", 0x83, " 0x", 0x84, " 0x", 0x85, " 0x", 0x86, " 0x", 0x87, CR, LF, 0, 0
	lds		temp, seccnt
	dec		temp
	sts		seccnt, temp
	brne	printids010
	sbi		PORTD, DS
	sbi		PORTD, MOT
	ret

indexx:
	cbi		PORTD, MOT
	cbi		PORTD, DS
	ldi		count, 0
	rcall	delayms				; wait 256ms
	ldi		count, 0
	rcall	delayms				; wait 256ms
	ldi		count, 15

indexx010:
	sbic	PIND, IDX
	rjmp	PC-1
	ldi		char, 'i'
	rcall	serout
	sbis	PIND, IDX
	rjmp	PC-1
	dec		count
	brne	indexx010
	sbi		PORTD, DS
	sbi		PORTD, MOT
	ret

;--------------------------------------------------------------------------
;
;	Delay 100usec + 1 cycles at 16MHz
;
delay100us:
;	rcall					;    3
	push	count			;    2
;
;	n * 15 cycles
;	
	ldi		count, 106		; 1590 cycles
delay100us010:
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	rjmp	PC+1
	dec		count
	brne	delay100us010
	pop		count			;   2
	ret						;   4
;
;
;
delay1ms:
	push	count
	ldi		count, 10
delay1ms010:
	rcall	delay100us
	dec		count
	brne	delay1ms010
	pop		count
	ret
;
;
;
delayms:
	rcall	delay1ms
	dec		count
	brne	delayms
	ret
;--------------------------------------------------------------------------
;
;
;	
idelay:
	cbi		PORTD, MOT			; Motor On
	cbi		PORTD, DS
	ldi		count, 0
	rcall	delayms				; wait 256ms
	ldi		count, 0
	rcall	delayms				; wait 256ms
	clr		xl
	clr		xh
iloop:	
	sbic	PINC, RD
	adiw	xh:xl, 1
	sbis	PINC, RD
	brne	iloop	
	sbi		PORTD, DS
	sbi		PORTD, MOT
;--------------------------------------------------------------------------
;
;
;	
seek:
	cbi		PORTD, MOT			; Motor On
	cbi		PORTD, DS
	ldi		count, 0
	rcall	delayms				; wait 256ms
	ldi		count, 0
	rcall	delayms				; wait 256ms

	lds		temp, a1l
	tst		temp				; Track 00
	breq	seekhome

	cpi		temp, 80
	brlo	seekok
	rjmp	seekdone

seekok:
	lds		r0, track			; r0 = current track
	sts		track, temp			; temp = desired track
	sts		pprint+0, r0
	sts		pprint+1, temp
	call	print
	.db		"Seek from 0x", 0x80, " to 0x", 0x81, CR, LF, 0x00, 0x00
	cp		temp, r0			
	brlo	seekout				; desired track < current track
	brne	seekin				; desired track > current track
	rjmp	seekdone			; already on track
;
;	desired track > current track
;
seekin:
	cbi		PORTD, DIR			; Low = towards centre
	rcall	delay100us
	cbi		PORTC, STEP
	rcall	delay100us
	sbi		PORTC, STEP	
	ldi		count, 6
	rcall	delayms
	inc		r0					; current track
	cp		temp, r0
	brne	seekin
	rjmp	seekdone
;
;	desired track < current track
;
seekout:

	sbi		PORTD, DIR			; High = away from centre
	rcall	delay100us
	cbi		PORTC, STEP
	rcall	delay100us
	sbi		PORTC, STEP	
	ldi		count, 6
	rcall	delayms
	dec		r0
	cp		temp, r0
	brne	seekout
	rjmp	seekdone

seekhome:
	sbi		PORTD, DIR
	rcall	delay100us
	cbi		PORTC, STEP
	rcall	delay100us
	sbi		PORTC, STEP	
	ldi		count, 6
	rcall	delayms
	sbic	PINC, TRK0
	rjmp	seekhome
	sts		track, zero

seekdone:
	sbi		PORTD, MOT
	sbi		PORTD, DS
	ret
	
;--------------------------------------------------------------------------
;
;	Read Track reads the address and data records of a diskette and not
;	the raw track. It assumes that an address section is always followed
;	by a data section (you could actually fool every computer by creating
;	a diskette which violates this assumption). Every time it encounters
;	an address section it adds the first 8 bytes read after the sync mark
;	to the trackbuffer and then reads 515 bytes following the next sync
;	mark and adds it to the trackbuffer and then alignes the buffer pointer
;	to the next 8 bits. It does this numsect times, assuming that we have a
;	HD Floppy in standard 1.44Mbyte format.
;

.equ	trackbuffer	= 0x1000		; where we place the buffer during tests
.equ	tracksize	= 0x2800		; size to initialise
.equ	sectorbuffer= 0x3800		; Used by the write sector routine

	.macro	btime
	sbis	PINC, RD				; Make sure we begin with RD = High
	inc		miss					; internal missed pulse counter
	sbi		PORTB, 0				; Debugging output
	sbic	PINC, RD				; Wait for a pulse to start
	rjmp	PC-1
	in		temp, TCNT0				; Get Timer value
	out		TCNT0, zero				; Reset Timer for next interval
	cbi		PORTB, 0				; Debugging output
	.endmacro

;
;	sbyte saves the character in the buffer (Y) reloads the bit counter 
;	with 8 and decrements the number of bytes requested (X). sbyte is
;	typically followed by a breq or brne instruction
;
;	At 16MHz the debugging output is creating misses but not read
;	faults. As we do no longer need the byte write debug we remove
;	this when running at 16Mhz. We 16MHz? Because we need a nice
;	frequency for the write routine. 
;
	.macro	sbyte
;	sbi		PORTB, 2				; Debugging output
	st		Y+, char				; Store byte in buffer
	ldi		count, 8				; Set bitcounter
;	cbi		PORTB, 2				; Debugging output
	sbiw	xh:xl, 1				; Subtract number of bytes to read
	.endmacro

;
;	Read Track using CPLD
;
quickread:
	cbi		PORTD, MOT
	cbi		PORTD, DS
	ldi		count, 0
	rcall	delayms				; wait 256ms
	ldi		count, 0
	rcall	delayms				; wait 256ms

	sbic	PIND, IDX
	rjmp	PC-1
	
	ldi		yl, low(trackbuffer)	; Buffer
	ldi		yh, high(trackbuffer)
	ldi		temp, 18				; 18 sectors
	sts		seccnt, temp
;
;
;
quicksectors:
	movw	zh:zl, yh:yl			; Save buffer
quicksectors010:	
	ldi		xl, low(8)				; Bytes to collect
	ldi		xh, high(8)
	clr		miss
	rcall	quicksection
	brcs	quickerror010
	ldd		temp, Z+0
	cpi		temp, 0xFE
	breq	quicksectors020
	movw	yh:yl, zh:zl
	rjmp	quicksectors010
	
quicksectors020:
	std		Z+7, miss				; Save the miss counter
	ldi		xl, low(515)
	ldi		xh, high(515)
	clr		miss
	rcall	quicksection
	brcs	quickerror020
	st		Y+, miss
	adiw	yh:yl, 7				; Align buffer address to a multiple of 8
	andi	yl, 0xF8
	lds		temp, seccnt
	dec		temp
	sts		seccnt, temp
	brne	quicksectors
	sbi		PORTD, DS
	sbi		PORTD, MOT
	ret

quickerror010:
	rjmp	readerror010
quickerror020:
	rjmp	readerror020
;
;
;
quicksection:
;
;	Y		Buffer Address
;	X		Number of Bytes to Read
;
;	Timer 1 is used to synchronise writes. Whenever we read a record
;	we start the timer as soon as we have seen the sync marks. The
;	initial value of timer 1 is set so that it overflows when we 
;	assume that the data record should be written. Not that we make
;	assumptions regarding the gaps used, in case we use different
;	floppy formats than DOS we need to adjust the initial value.
;	We use a timer so we do not need to count the cycles we spend
;	from the sync mark to the real write data record
;
;	Reset timer
;
	sts		TCCR1B, zero		; Stop Timer
	sts		TCCR1A, zero		; No Compare Mode, Normal Counter Mode
	sbi		TIFR1, TOV1			; Clear timer overflow bit
;
;	We use counter 1 in free running mode and with the clock divided
;	by 8. Each byte in MFM requires 16usec so we need to need to wait
;	about 430usec before we start to write, so we need to setup the
;	initial timer with a value that sets the overflow bit after
;
	ldi		temp, high(-F_CPU/8*16*27/1000000)
	sts		TCNT1H, temp
	ldi		temp, low(-F_CPU/8*16*27/1000000)
	sts		TCNT1L, temp
;
;	Ack data and Reset statemachine. Once the state machine is reset
;	it will only signal a byte when it encounters the second sync
;	mark. Note the statemachine already detects the first sync mark
;	but does not make it available to the MCU.
;
	ldi		count, 3
	cbi		PORTB, 0
quickretry:
	sbi		PORTB, ACK
	cbi		PORTB, RES
	cbi		PORTB, ACK			;
	sbi		PORTB, RES			;

	sbis	PINB, DRQ
	rjmp	PC-1
	in		temp, PINA
	sbi		PORTB, ACK
	cbi		PORTB, ACK			; Ack first byte
	cpi		temp, 0xA1			; Must be sync
	brne	quicksectionerr
	
	sbis	PINB, DRQ
	rjmp	PC-1
	in		temp, PINA
	sbi		PORTB, ACK
	cbi		PORTB, ACK			; Ack second byte
	cpi		temp, 0xA1			; Must be sync
	brne	quicksectionerr


	ldi		temp, (1<<CS11)
	sts		TCCR1B, temp			; Start Counter, prescaler 8

;
;	Read Loop to retrieve the bytes from the CPLD. The CPLD sets
;	DRQ when a byte is ready to be read. The MCU reads the byte
;	and resets DRQ (via ACK) so it knows when another byte is ready.
;	For a HD floppy we get a byte every 16usec. For a MFM HD this
;	is every 1.6usec so this routine must have less than 25 cycles
;	in the worst case (note that when we just miss DRQ to ready 
;	loop that checks DRQ takes another 4 cycles.
;

quicksection010:				;
	sbis	PINB, DRQ				;  DATA8
	rjmp	PC-1				;	 4
	in		temp, PINA			;	 1
	st		Y+, temp			;	 2
	sbi		PORTB, ACK			; 	 2
	cbi		PORTB, ACK			;	 2
	sbiw	X, 1				;	 2
	brne	quicksection010		;	 2
								;	==
								;	15
	clc
	ret
	
quicksectionerr:
;	dec		count
;	brne	quickretry
	sbi		PORTB, 0
	cbi		PORTB, 0
	sbi		PORTB, 0
	cbi		PORTB, 0
	sbi		PORTB, 0
	cbi		PORTB, 0
	sec
	ret
	
;--------------------------------------------------------------------------
;
;	Read Track without hardware support using only the IO pins
;
;
readtrack:
	cbi		PORTD, MOT
	cbi		PORTD, DS
	ldi		count, 0
	rcall	delayms				; wait 256ms
	ldi		count, 0
	rcall	delayms				; wait 256ms
;
;	Initialise
;
	out		TCCR0A, zero
	out		TCCR0B, one
	out		TCNT0, zero
	ldi		xl, low(tracksize)		; Bytes to collect
	ldi		xh, high(tracksize)
	ldi		yl, low(trackbuffer)	; Buffer
	ldi		yh, high(trackbuffer)

	ldi		temp, 0xa5				; First fill our buffer
readfill:							; with a pattern
	st		Y+, temp
	sbiw	xh:xl, 1
	brne	readfill

	ldi		yl, low(trackbuffer)	; Buffer
	ldi		yh, high(trackbuffer)
	ldi		temp, numsect			; Set number of sectors to read
	sts		seccnt, temp			; 

	sbic	PIND, IDX
	rjmp	PC-1

;
;
;
readsectors:
;
;	First we just ask for 8 bytes of a section. Section is
;	the data that immediately follows the sync mark, this 
;	can either be an address or a data block. Note that there
;	is a rather large GAP between each section. We have
;	enough time to do some processing without missing
;	any consecutive section. This is important as we need
;	to read the section immediately following the ID section
;	to be able to tell wich data section is which sector
;
	movw	zh:zl, yh:yl			; Save buffer
readsectors000:	
	ldi		xl, low(8)				; Bytes to collect
	ldi		xh, high(8)
	clr		miss
	rcall	readsection				; Read (8 bytes) from next record on floppy
	brcs	readsectors005
	ldd		temp, Z+0				; Get first byte
	cpi		temp, 0xFE				; is it an address mark
	breq	readsectors010			; yes found a sector address
readsectors005:
	movw	yh:yl, zh:zl			; Restore buffer, next section is
	rjmp	readsectors000			; most likely an address section

readsectors010:
	std		Z+7, miss				; Save the miss counter
;
;	Now we assume that the section following the address section is
;	a data section. We must read exactly 3 bytes more than the sector
;	length. That is the mark, the data and the two CRC bytes. It is not
;	guaranteed that all pulses following the CRC are valid MFM pulses.
;
	ldi		xl, low(128)
	ldi		xh, high(128)			; Assume 128bytes sectors
	ldd		temp, Z+4				; Sector Length
	dec		temp
	brmi	readsectors020			; Done
	add		xl, xl
	adc		xh, xh					; Assume 256bytes sectors
	dec		temp
	brmi	readsectors020			; Done
	add		xl, xl
	adc		xh, xh					; Assume 512bytes sectors
	dec		temp
	brmi	readsectors020			; Done
	add		xl, xl
	adc		xh, xh					; Else we have 1024bytes sectors
readsectors020:
	adiw	xh:xl, 3				; Address Mark, CRC
	clr		miss
	rcall	readsection
	brcs	readerror020
	st		Y+, miss
	adiw	yh:yl, 7				; Align buffer address to a multiple of 8
	andi	yl, 0xF8
	lds		temp, seccnt			; More sectors to do
	dec		temp
	sts		seccnt, temp
	brne	readsectors				; Yes
	
readexit:
	sbi		PORTD, MOT
	sbi		PORTD, DS
	ret	

readerror010:
	sts		pprint+0, yl
	sts		pprint+1, yh
	sts		pprint+2, miss
	call	print
	.db		"Error reading address section 0x", 0x81, 0x80, " miss 0x", 0x82, CR, LF, 0
	rjmp	readexit

readerror020:
	sts		pprint+0, yl
	sts		pprint+1, yh
	sts		pprint+2, miss
	call	print
	.db		"Error reading data section 0x", 0x81, 0x80, " miss 0x", 0x82,  CR, LF, 0, 0
	rjmp	readexit


;--------------------------------------------------------------------------
;
;	Read a section from the diskette. First we look for the synchronisation
;	Gap (12 x 0x00) and then check for three SYNC MARKS. After the sync
;	marks we start timer 1 and read and decode the MFM bitstream we will
;	read X bytes into the buffer at Y
;
;	Y		Buffer Address
;	X		Number of Bytes to Read
;
readsection:

;
;	Timer 1 is used to synchronise a sector write. This is because the
;	write routine must start to write the data record after the GAP
;	which follows the ID record. The ID record has the following 
;	structure
;
;	0xFB, TRK, HD, SEC, LEN, CRCH, CRCL, 22*0x4e(GAP)
;
;	For this we initialise the timer counter with a value that lets
;	the timer overflow when the head has reached the end of the GAP
;	and start the timer right after we have found the SYNC MARKS.
;	
;	The write sector routine then only needs to read the ID sections
;	and if the correct ID section has been found it only needs to 
;	check the timer overflow to know when it needs to begin writing
;	the data record.
;
	sts		TCCR1B, zero		; Stop Timer
	sts		TCCR1A, zero		; No Compare Mode, Normal Counter Mode
	sbi		TIFR1, TOV1			; Clear timer overflow bit
;
;	We use counter 1 in free running mode and with the clock divided
;	by 8. Each byte in MFM requires 16usec so we need to need to wait
;	about 430usec before we start to write, so we need to setup the
;	initial timer with a value that sets the overflow bit after
;
	ldi		temp, high(-F_CPU/8*16*27/1000000)
	sts		TCNT1H, temp
	ldi		temp, low(-F_CPU/8*16*27/1000000)
	sts		TCNT1L, temp

	
;
;	Read at least 80 consecutive short pulses (normally there are 96
;	pulses from the 12 x 0x00 preamble), we could also read less pulses
;	this has no effect on the sync mark detection, we could as well try
;	to find the sync marks. But I leave them here for nostalgic reasons.
;
;	......SSSSSSSSSSSSSS
;
sstart:
	ldi		count, 80				; less than 12*8
sloop:
	btime
	cpi		temp, t_short
	brsh	sstart					; Not a short pulse
	dec		count
	brne	sloop					; Until we have 80 consecutive short pulses
;
;	Between the consecutive short pulses and the first SYNC MARK we expect
;	a medium pulse
;
sskip:
	btime
	cpi		temp, t_short
	brlo	sskip					; Another short pulse is ok
	cpi		temp, t_long
	brlo	wlong1					; Yes a medium pulse
	rjmp	sstart					; Not what we expected 
;
;	Now we start to find sync marks, there are 3 consecutive SYNC MARKS
;	Sync Marks are the value 0xA1 written with a missing clock pulse without
;	violating the MFM rules regarding the distance between flux changes. This
;	produces the following interval pattern 
;
;	....LMLMSLMLMSLMLM
;
;			    sync mark	    sync mark	    sync mark
;	Data	1 0 1 0 0 0 0 1| 1 0 1 0 0 0 0 1| 1 0 1 0 0 0 0 1|
;	Clock	 0 0 0 1 1 1 0 |0 0 0 0 1 1 1 0 |0 0 0 0 1 1 1 0 |
;	MFM		100010010001001|0100010010001001|0100010010001001|
;                    ^                ^                ^
;	Note missing clock in MFM
;
;	We will only continue if the pattern of interval matches, else we will
;	start from the beginning (sstart == sync start)
;
wlong1:
	btime					; L
	cpi		temp, t_long
	brlo	wlong11
;
	btime					; M
	cpi		temp, t_short
	brlo	wlong11
	cpi		temp, t_long
	brsh	wlong11
;
	btime					; L
	cpi		temp, t_long
	brlo	wlong11
;
	btime					; M
	cpi		temp, t_short
	brlo	wlong11
	cpi		temp, t_long
	brsh	wlong11
;
	btime					; S
	cpi		temp, t_short
	brlo	wlong2
;
;	Jump label in case of no match
;
wlong11:
	rjmp	sstart
;	
wlong2:
	btime					; L
	cpi		temp, t_long
	brlo	wlong11
;
	btime					; M
	cpi		temp, t_short
	brlo	wlong11
	cpi		temp, t_long
	brsh	wlong11
;
	btime					; L
	cpi		temp, t_long
	brlo	wlong11
;
	btime					; M
	cpi		temp, t_short
	brlo	wlong11
	cpi		temp, t_long
	brsh	wlong11
;
	btime					; S
	cpi		temp, t_short
	brlo	wlong3
;
;	Jump label in case of no match
;
wlong22:
	rjmp	sstart
;
wlong3:
	btime					; L
	cpi		temp, t_long
	brlo	wlong22
;
	btime					; M
	cpi		temp, t_short
	brlo	wlong22
	cpi		temp, t_long
	brsh	wlong22
;
	btime					; L
	cpi		temp, t_long
	brlo	wlong22
;
	btime					; M
	cpi		temp, t_short
	brlo	wlong22
	cpi		temp, t_long
	brsh	wlong22
;
;	After the last btime macro call we only have about 1usec of processing
;	that is we should not spend more then 16 cycels before the next btime
;	macro call. 
;
	clr		miss					; Reset miss counter
	ldi		temp, (1<<CS11)
	sts		TCCR1B, temp			; Start Counter, prescaler 8
	ldi		count, 8				; Initialise bit counter
	
;
;	we spotted three sync marks now we need to analyze further gaps
;	depending on the current state and the interval size we switch
;	states (l00x0). We collect bits in the register named "char"
;	register "count" is the bitcounter, whenever we have accumulated
;	8 bits we need to store the "char" in the buffer and reload the
;	bitcounter. 
;
;	Here we are in phase with a "data" pulse
;
l0010:															; 
	btime														;  8
	cpi		temp, t_short										;  1
	brsh	l0020												;  2	-> 11
;---------------------------------------------------------
;
;	Short Interval. Next Bit is a '1'
;			--------------->
;	Data	1 0 1 0 0 0 0 1| 1
;	Clock	 0 0 0 1 1 1 0 |0
;	MFM		100010010001001|01
;
	sec															;
	rol		char												; We have a '1'
	dec		count												; We got one more bit
	brne	l0010												; Still more to go
	sbyte														; Save byte, reload bitcounter and decrement number of bytes requested
	brne	l0010												; Need more bytes
	rjmp	readsectiondone										; finished

l0020:
	cpi		temp, t_long										;  1
	brlo	l0040												;  2	-> 14
;---------------------------------------------------------
;
;	Long Interval. Next are two bits '01'
;			--------------->
;	Data	1 0 1 0 0 0 0 1| 0 1
;	Clock	 0 0 0 1 1 1 0 |0 0
;	MFM		100010010001001|0001
;
	lsl		char												;
	dec		count												; we need to check the number of collected bits after every single bit added!
	brne	l0030												;
	sbyte														;
	brne	l0030												;
	rjmp	readsectiondone										;
l0030:
	sec															;
	rol		char												;
	dec		count												;
	brne	l0010												;
	sbyte
	brne	l0010												;
	rjmp	readsectiondone
;---------------------------------------------------------
;
;	Note section l0060 moved to here so we can avoid 
;	the intermediate label l0070 but reach l0010 directly
;	(see archived versions)
;
;	Here we are in phase with a "clock" pulse
;
;	last pulse was on a clock and the next data bit is either
;	a zero or a one in which case the next gap is either
;	a short (zero) or a medium(one) a long gap at this
;	stage would be equivalent to a sync mark or a violation
;	violations can occur after the CRC of a data field as
;	writing of data stops right after the CRC and due to
;	various reasons this is not necesserily in sync with
;	the GAP pattern on the diskette.
;
l0060:
	btime
	cpi		temp, t_short
	brlo	l0080
	cpi		temp, t_long
	brsh	l0090
;---------------------------------------------------------
;
;	we have a medium gap and therefore the data bit is a '1'
;	and we are again aligned to the data pulse
;
;			------------------>
;	Data	1 0 1 0 0 0 0 1| 0 0 1
;	Clock	 0 0 0 1 1 1 0 |0 1 0
;	MFM		100010010001001|001001
;
	sec
	rol		char
	dec		count
	brne	l0010
	sbyte
	brne	l0010
	rjmp	readsectiondone
;---------------------------------------------------------
;
;	Here we are in phase with a "data" pulse
;	
;	1 1/2 bit cells so we have a clock pulse and next two
;	bits are '00'
;			--------------->
;	Data	1 0 1 0 0 0 0 1| 0 0
;	Clock	 0 0 0 1 1 1 0 |0 1 
;	MFM		100010010001001|001
;
l0040:
	lsl		char												;  1
	dec		count												;  1
	brne	l0050												;  2	-> 18
	sbyte
	brne	l0050
	rjmp	readsectiondone
;
l0050:
	lsl		char												;  1
	dec		count												;  1
	brne	l0060												;  1
	sbyte
	brne	l0060												;  2	-> 28
	rjmp	readsectiondone
;---------------------------------------------------------
;
;	Here we are in phase with a "clock" pulse
;
;	we have a short pulse, and therefore the data bit is
;	zero and we are back at the situation where the pulse
;	was on a clock
;
;			------------------>
;	Data	1 0 1 0 0 0 0 1| 0 0 0
;	Clock	 0 0 0 1 1 1 0 |0 1 1
;	MFM		100010010001001|00101
;
l0080:
	lsl		char
	dec		count
	brne	l0060
	sbyte
	brne	l0060
	rjmp	readsectiondone
;---------------------------------------------------------
;
;	Here we are in phase with a "clock" pulse and at the
;	same time we have a long gap, at this stage this is 
;	equivalent to a missing clock pulse.
;
;	Note: A long gap can normally not occur with the
;	pulses synced to the clock position. Such a gap
;	would require that the data bit after the first
;	pulse is 0 and at the same time the next data bit
;	which is right before the second pulse as well. 
;	But two data bits produce a clock pulse in between
;	which obviously is missing (see X) as such this
;	is considered to be a read error
;
;			------------------>
;	Data	1 0 1 0 0 0 0 1| 0 0 0 0
;	Clock	 0 0 0 1 1 1 0 |0 1 X 1
;	MFM		100010010001001|0010001
;

l0090:
	sec
	ret
	
readsectiondone:
	clc
	ret

;--------------------------------------------------------------------------
;
;
;
analyzetrack:
	call	print
	.db		"Buff    Mark Trk  Side Sect Leng HCRC   HCRC(c) CRC     CRC(c)  Miss  Miss", CR, LF, 0, 0
	ldi		temp, numsect
	sts		seccnt, temp

	ldi		yl, low(trackbuffer)
	ldi		yh, high(trackbuffer)
analyze000:
	mov		crcl, ff
	mov		crch, ff
	ldi		temp, 0xa1
	updcrc	temp
	ldi		temp, 0xa1
	updcrc	temp
	ldi		temp, 0xa1
	updcrc	temp
	
	ld		temp, Y+				; Address Mark
	sts		pprint+0, temp
	updcrc	temp
	ld		temp, Y+				; Track
	sts		pprint+1, temp
	updcrc	temp
	ld		temp, Y+				; Side
	sts		pprint+2, temp
	updcrc	temp
	ld		temp, Y+				; Sector
	sts		pprint+3, temp
	updcrc	temp
	ld		temp, Y+				; Length
	mov		count, temp				; we need it
	sts		pprint+4, temp
	updcrc	temp
	ld		temp, Y+				; HCRCL
	sts		pprint+5, temp
	ld		temp, Y+				; HCRCH
	sts		pprint+6, temp
	sts		pprint+7, crcl
	sts		pprint+8, crch
	ld		temp, Y+
	sts		pprint+13, temp			; misses
	sts		pprint+14, yl
	sts		pprint+15, yh
;
;	Count = sector size/4
;
	mov		temp, count
	ldi		count, 128/4
	dec		temp
	brmi	analyze005
	ldi		count, 256/4
	dec		temp
	brmi	analyze005
	ldi		count, 512/4
	dec		temp
	brmi	analyze005
	clr		count					; count=0 corresponds to 1024bytes
analyze005:	
	mov		crcl, ff
	mov		crch, ff
	ldi		temp, 0xa1
	updcrc	temp
	ldi		temp, 0xa1
	updcrc	temp
	ldi		temp, 0xa1
	updcrc	temp
	ld		temp, Y+
	updcrc	temp
analyze010:
	ld		temp, Y+				; Do four bytes (count = sectorsize/4)
	updcrc	temp
	ld		temp, Y+
	updcrc	temp
	ld		temp, Y+
	updcrc	temp
	ld		temp, Y+
	updcrc	temp
	dec		count

	brne	analyze010
	ld		temp, Y+
	sts		pprint+9, temp
	ld		temp, Y+
	sts		pprint+10, temp
	sts		pprint+11, crcl
	sts		pprint+12, crch
	call	print
	.db		"0x", 0x8f, 0x8e, "  "
	.db		"0x", 0x80, " ", "0x", 0x81, " "
	.db		"0x", 0x82, " ", "0x", 0x83, " "
	.db		"0x", 0x84, " ", "0x", 0x85, 0x86, " ", "0x", 0x88, 0x87, "  0x", 0x89, 0x8a, "  0x", 0x8c, 0x8b, "  0x", 0x8d, "  0x", 0, 0
	ld		temp, Y+
	sts		pprint+0, temp			; miss
	call	print
	.db		0x80, CR, LF, 0

	adiw	yh:yl, 7
	andi	yl, 0xF8
	lds		temp, seccnt
	dec		temp
	sts		seccnt, temp
	breq	analyzedone
	rjmp	analyze000

analyzedone:
	ret

;
;--------------------------------------------------------------------------
;
;	AVR monitor. A interactive monitor inspired by the Apple II
;
.include	"../include/monitor-v1-6.asm"

.include "MFMWriter.asm"

;--------------------------------------------------------------------------
;
;	Monitor Function Routines
;
helpscreen:
	lds		temp, a1l
	cpi		temp, 0x01
	brne	helpscreen2
	ldi		zl, low(2*help1)
	ldi		zh, high(2*help1)
	jmp		splashsend
helpscreen2:
	cpi		temp, 0x02
	brne	helpscreen3
	ldi		zl, low(2*help2)
	ldi		zh, high(2*help2)
	jmp		splashsend
helpscreen3:
	cpi		temp, 0x03
	brne	helpscreen4
	ldi		zl, low(2*help3)
	ldi		zh, high(2*help3)
	jmp		splashsend
helpscreen4:
	cpi		temp, 0x04
	brne	helpscreen5
	ldi		zl, low(2*help4)
	ldi		zh, high(2*help4)
	jmp		splashsend
helpscreen5:						; Default Help Screen
	ldi		zl, low(2*hello)
	ldi		zh, high(2*hello)
	jmp		splashsend

hello:	.db	0x0d, 0x0a, 0x0d, 0x0a
;		"----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8"
	.db	"MFM Reader/Writer Test Environment Vers 0.2  ",__DATE__, "                      ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"     For general syntax see Apple II monitor. The following commands of the     ", 0x0d, 0x0a
	.db	"     Apple II Monitor are implemented: blank, dot, colon, -, +, <, move, verify ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"     ?        Show Helpscreen                                                   ", 0x0d, 0x0a
	.db	"     1?       Help for read/write using the software only routines              ", 0x0d, 0x0a
	.db	"     2?       Help for read/write using the hardware assisted routines          ", 0x0d, 0x0a
	.db	"     3?       General information                                               ", 0x0d, 0x0a
	.db	"     4?       Description of the report produced by the K command               ", 0x0d, 0x0a
	.db	"     nnT      Seek to track nn, n=0 will do a home                              ", 0x0d, 0x0a
	.db	"     K        Analyses the track read into memory and prints some info          ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"     Currently the routine supports just 3 1/2 HD floppies and assumes that     ", 0x0d, 0x0a
	.db	"     each track has 18 sectors with 512bytes each. It's a simple task to        ", 0x0d, 0x0a
	.db	"     adapt to other formats.                                                    ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"     The read track routines (see other help screens) use memory starting       ", 0x0d, 0x0a
	.db	"     from 0x1000. You can use the Apple II like commands to inspect the         ", 0x0d, 0x0a
	.db	"     result.                                                                    ", 0x0d, 0x0a
	.db	"     The write sector routines use the memory from 0x3800 as a temporary        ", 0x0d, 0x0a
	.db	"     buffer.                                                                    ", 0x0d, 0x0a
	.dw	0
;
help1:	.db	0x0d, 0x0a, 0x0d, 0x0a
;		"----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8"
	.db	"MFM Reader/Writer Test Environment Vers 0.2  ",__DATE__, "                      ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"     Software only floppy access routines                                       ", 0x0d, 0x0a
	.db	"     R        Reads the current track of side 0 of the floppy to memory         ", 0x0d, 0x0a
	.db	"              starting at 0x1000. It will store 8 bytes from the address        ", 0x0d, 0x0a
	.db	"              record followd by 520 of the data record. The record in memory    ", 0x0d, 0x0a
	.db	"              include the mark and CRC. Use the command K to print a report.    ", 0x0d, 0x0a
	.db	"     s<nnW    This will write the buffer at memory address nn to sector s       ", 0x0d, 0x0a
	.db	"              it is recommended to use only buffers below 0x1000                ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"     Note: make sure that WD from the AVR is connected to the floppy drive      ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.dw	0
;
help2:	.db	0x0d, 0x0a, 0x0d, 0x0a
;		"----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8"
	.db	"MFM Reader/Writer Test Environment Vers 0.2  ",__DATE__, "                      ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"     Hardware assisted floppy access routines                                   ", 0x0d, 0x0a
	.db	"     Q        Reads the current track of side 0 of the floppy to memory         ", 0x0d, 0x0a
	.db	"              starting at 0x1000. It will store 8 bytes from the address        ", 0x0d, 0x0a
	.db	"              record followd by 520 of the data record. The record in memory    ", 0x0d, 0x0a
	.db	"              include the mark and CRC. Use the command K to print a report.    ", 0x0d, 0x0a
	.db	"     s<nnU    This will write the buffer at memory address nn to sector s       ", 0x0d, 0x0a
	.db	"              it is recommended to use only buffers below 0x1000                ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"     Note: make sure that WD from the CPLD is connected to the floppy drive     ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.dw	0
;
help3:	.db	0x0d, 0x0a, 0x0d, 0x0a
;		"----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8"
	.db	"MFM Reader/Writer Test Environment Vers 0.2  ",__DATE__, "                      ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"     This monitor program has a command line similar to the Apple II monitor    ", 0x0d, 0x0a
	.db	"     program. All numbers and addresses are hexadecimal. It implements the      ", 0x0d, 0x0a
	.db	"     commands . < : + - M V and <CR> of the Apple II monitor for more infor-    ", 0x0d, 0x0a
	.db	"     mation see the Monitor Chapter of a Apple II manual                        ", 0x0d, 0x0a
	.db	"     In addition the these standard commands it als implements the following    ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"     aaaa.bbbbP   This will store the pattern 0x00, 0x01, 0x02, ... to the      ", 0x0d, 0x0a
	.db	"                  specified memory range                                        ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"     X            Will print 16 bytes from the current address aligned          ", 0x0d, 0x0a
	.db	"                  to a 16 byte boundaryin the same format as the unix           ", 0x0d, 0x0a
	.db	"                  hexdump -C command                                            ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"     S            Toggle Drive Select                                           ", 0x0d, 0x0a
	.db	"     O            Toggle Motor On                                               ", 0x0d, 0x0a
	.db	"     H            Toggle Head Select                                            ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.dw	0
;
help4:	.db	0x0d, 0x0a, 0x0d, 0x0a
;		"----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8"
	.db	"MFM Reader/Writer Test Environment Vers 0.2  ",__DATE__, "                      ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"     The K command produces some output with the following column. It assuems   ", 0x0d, 0x0a
	.db	"     that a track been read using either the R or Q command and placed at       ", 0x0d, 0x0a
	.db	"     memory location 0x1000 with 8bytes address and 520bytes data records.      ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"   Buff    Mark Trk  Side Sect Leng HCRC   HCRC(c) CRC     CRC(c)  Miss  Miss   ", 0x0d, 0x0a
	.db	"   0x1008  0xFE 0x45 0x00 0x09 0x02 0x911F 0x911F  0x2BF6  0x2BF6  0x00  0x00   ", 0x0d, 0x0a
	.db	"   0x1218  0xFE 0x45 0x00 0x0A 0x02 0xC44C 0xC44C  0x2BF6  0x2BF6  0x00  0x00   ", 0x0d, 0x0a
	.db	"   0x1428  0xFE 0x45 0x00 0x0B 0x02 0xF77D 0xF77D  0x2BF6  0x2BF6  0x00  0x00   ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.db	"     Buff    The address of the assumed data record starting with the mark.     ", 0x0d, 0x0a
	.db	"     Mark    and following show the address record                              ", 0x0d, 0x0a
	.db	"     HCRC    The CRC of the address record as read from the floppy              ", 0x0d, 0x0a
	.db	"     HCRC(c) The CRC as expected (calculated by the program)                    ", 0x0d, 0x0a
	.db	"     CRC     The CRC of the data record as read from the floppy                 ", 0x0d, 0x0a
	.db	"     CRC(c)  The CRC of the data record as expected/calculated                  ", 0x0d, 0x0a
	.db	"     MISS    Number of misses when reading the pulses for address and           ", 0x0d, 0x0a
	.db	"             data record must be 0x00 else we have a serious issue              ", 0x0d, 0x0a
	.db	"                                                                                ", 0x0d, 0x0a
	.dw	0

	.include	"../include/print-v1-1.asm"

	align	8
;=========================================================================
;
;	Note the updcrc macro assumes that the tables are page aligned
;
;	low byte CRC lookup table
; 
	align	8
crclo:
 .db 0x00,0x21,0x42,0x63,0x84,0xA5,0xC6,0xE7,0x08,0x29,0x4A,0x6B,0x8C,0xAD,0xCE,0xEF
 .db 0x31,0x10,0x73,0x52,0xB5,0x94,0xF7,0xD6,0x39,0x18,0x7B,0x5A,0xBD,0x9C,0xFF,0xDE
 .db 0x62,0x43,0x20,0x01,0xE6,0xC7,0xA4,0x85,0x6A,0x4B,0x28,0x09,0xEE,0xCF,0xAC,0x8D
 .db 0x53,0x72,0x11,0x30,0xD7,0xF6,0x95,0xB4,0x5B,0x7A,0x19,0x38,0xDF,0xFE,0x9D,0xBC
 .db 0xC4,0xE5,0x86,0xA7,0x40,0x61,0x02,0x23,0xCC,0xED,0x8E,0xAF,0x48,0x69,0x0A,0x2B
 .db 0xF5,0xD4,0xB7,0x96,0x71,0x50,0x33,0x12,0xFD,0xDC,0xBF,0x9E,0x79,0x58,0x3B,0x1A
 .db 0xA6,0x87,0xE4,0xC5,0x22,0x03,0x60,0x41,0xAE,0x8F,0xEC,0xCD,0x2A,0x0B,0x68,0x49
 .db 0x97,0xB6,0xD5,0xF4,0x13,0x32,0x51,0x70,0x9F,0xBE,0xDD,0xFC,0x1B,0x3A,0x59,0x78
 .db 0x88,0xA9,0xCA,0xEB,0x0C,0x2D,0x4E,0x6F,0x80,0xA1,0xC2,0xE3,0x04,0x25,0x46,0x67
 .db 0xB9,0x98,0xFB,0xDA,0x3D,0x1C,0x7F,0x5E,0xB1,0x90,0xF3,0xD2,0x35,0x14,0x77,0x56
 .db 0xEA,0xCB,0xA8,0x89,0x6E,0x4F,0x2C,0x0D,0xE2,0xC3,0xA0,0x81,0x66,0x47,0x24,0x05
 .db 0xDB,0xFA,0x99,0xB8,0x5F,0x7E,0x1D,0x3C,0xD3,0xF2,0x91,0xB0,0x57,0x76,0x15,0x34
 .db 0x4C,0x6D,0x0E,0x2F,0xC8,0xE9,0x8A,0xAB,0x44,0x65,0x06,0x27,0xC0,0xE1,0x82,0xA3
 .db 0x7D,0x5C,0x3F,0x1E,0xF9,0xD8,0xBB,0x9A,0x75,0x54,0x37,0x16,0xF1,0xD0,0xB3,0x92
 .db 0x2E,0x0F,0x6C,0x4D,0xAA,0x8B,0xE8,0xC9,0x26,0x07,0x64,0x45,0xA2,0x83,0xE0,0xC1
 .db 0x1F,0x3E,0x5D,0x7C,0x9B,0xBA,0xD9,0xF8,0x17,0x36,0x55,0x74,0x93,0xB2,0xD1,0xF0 
;
;	hi byte CRC lookup table
;
crchi:
 .db 0x00,0x10,0x20,0x30,0x40,0x50,0x60,0x70,0x81,0x91,0xA1,0xB1,0xC1,0xD1,0xE1,0xF1
 .db 0x12,0x02,0x32,0x22,0x52,0x42,0x72,0x62,0x93,0x83,0xB3,0xA3,0xD3,0xC3,0xF3,0xE3
 .db 0x24,0x34,0x04,0x14,0x64,0x74,0x44,0x54,0xA5,0xB5,0x85,0x95,0xE5,0xF5,0xC5,0xD5
 .db 0x36,0x26,0x16,0x06,0x76,0x66,0x56,0x46,0xB7,0xA7,0x97,0x87,0xF7,0xE7,0xD7,0xC7
 .db 0x48,0x58,0x68,0x78,0x08,0x18,0x28,0x38,0xC9,0xD9,0xE9,0xF9,0x89,0x99,0xA9,0xB9
 .db 0x5A,0x4A,0x7A,0x6A,0x1A,0x0A,0x3A,0x2A,0xDB,0xCB,0xFB,0xEB,0x9B,0x8B,0xBB,0xAB
 .db 0x6C,0x7C,0x4C,0x5C,0x2C,0x3C,0x0C,0x1C,0xED,0xFD,0xCD,0xDD,0xAD,0xBD,0x8D,0x9D
 .db 0x7E,0x6E,0x5E,0x4E,0x3E,0x2E,0x1E,0x0E,0xFF,0xEF,0xDF,0xCF,0xBF,0xAF,0x9F,0x8F
 .db 0x91,0x81,0xB1,0xA1,0xD1,0xC1,0xF1,0xE1,0x10,0x00,0x30,0x20,0x50,0x40,0x70,0x60
 .db 0x83,0x93,0xA3,0xB3,0xC3,0xD3,0xE3,0xF3,0x02,0x12,0x22,0x32,0x42,0x52,0x62,0x72
 .db 0xB5,0xA5,0x95,0x85,0xF5,0xE5,0xD5,0xC5,0x34,0x24,0x14,0x04,0x74,0x64,0x54,0x44
 .db 0xA7,0xB7,0x87,0x97,0xE7,0xF7,0xC7,0xD7,0x26,0x36,0x06,0x16,0x66,0x76,0x46,0x56
 .db 0xD9,0xC9,0xF9,0xE9,0x99,0x89,0xB9,0xA9,0x58,0x48,0x78,0x68,0x18,0x08,0x38,0x28
 .db 0xCB,0xDB,0xEB,0xFB,0x8B,0x9B,0xAB,0xBB,0x4A,0x5A,0x6A,0x7A,0x0A,0x1A,0x2A,0x3A
 .db 0xFD,0xED,0xDD,0xCD,0xBD,0xAD,0x9D,0x8D,0x7C,0x6C,0x5C,0x4C,0x3C,0x2C,0x1C,0x0C
 .db 0xEF,0xFF,0xCF,0xDF,0xAF,0xBF,0x8F,0x9F,0x6E,0x7E,0x4E,0x5E,0x2E,0x3E,0x0E,0x1E 

















